"""
Implement YAML document Merger.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
import sys
from os.path import basename
from typing import Any, Dict, List, Set, Tuple, Union
import json
from io import StringIO
from pathlib import Path

from ruamel.yaml.comments import (
    CommentedMap, CommentedSet, CommentedSeq, TaggedScalar
)

from yamlpath.common import Anchors, Nodes, Parsers
from yamlpath.wrappers import ConsolePrinter, NodeCoords
from yamlpath.merger.exceptions import MergeException
from yamlpath.merger.enums import (
    AnchorConflictResolutions,
    AoHMergeOpts,
    ArrayMergeOpts,
    HashMergeOpts,
    OutputDocTypes,
    SetMergeOpts,
)
from yamlpath.merger import MergerConfig
from yamlpath import YAMLPath, Processor


class Merger:
    """Performs YAML document merges."""

    DEPRECATION_WARNING = (
        "WARNING:  Deprecated methods will be removed in the next major"
        " release of yamlpath.  Please refer to the CHANGES file for more"
        " information (and how to get rid of this message).")
    depwarn_printed = False

    def __init__(
        self, logger: ConsolePrinter, lhs: Any, config: MergerConfig
    ) -> None:
        """
        Instantiate this class into an object.

        Parameters:
        1. logger (ConsolePrinter) Instance of ConsoleWriter or subclass
        2. lhs (Any) The prime left-hand-side parsed YAML data
        3. config (MergerConfig) User-defined document merging rules

        Returns:  N/A

        Raises:  N/A
        """
        self.logger: ConsolePrinter = logger
        self.config: MergerConfig = config
        self.data: Any = lhs

        # ryamel.yaml unfortunately tracks comments AFTER each YAML node.  As
        # such, it is impossible to copy comments from RHS to LHS in any
        # sensible way.  Trying leads to absurd merge results that are data-
        # accurate but comment-insane.  This ruamel.yaml design decision forces
        # me to simply delete all comments from all merge documents to produce
        # a sensible result.  That said, enable users to attempt to preserve
        # LHS comments.
        if not self.config.is_preserving_lhs_comments():
            Parsers.delete_all_comments(self.data)

    @property
    def data(self) -> Any:
        """Document data being merged into (accessor)."""
        return self._data

    @data.setter
    def data(self, value: Any) -> None:
        """Document data being merged into (mutator)."""
        if not self.config.is_preserving_lhs_comments():
            Parsers.delete_all_comments(value)
        self._data = value

    def _delete_mergeref_keys(self, data: CommentedMap) -> None:
        """
        Delete all YAML merge reference keys from a CommentedMap.

        This is necessary when using the insert() method of a CommentedMap
        because doing so converts all YAML merge references (key-value pairs
        merged into a YAML Hash via the `<<:` operator) to concrete key-value
        pairs of the affected Hash.
        """
        concrete_keys = []
        for local_key, _ in data.non_merged_items():
            concrete_keys.append(local_key)

        reference_keys = set(data.keys()).difference(concrete_keys)
        for key in reference_keys:
            self.logger.debug(
                "Deleting key from LHS:",
                data=key, prefix="Merger::_delete_mergeref_keys:  ",
                header="!" * 50
            )
            del data[key]

    #pylint: disable=too-many-branches,too-many-statements
    def _merge_dicts(
        self, lhs: CommentedMap, rhs: CommentedMap, path: YAMLPath
    ) -> CommentedMap:
        """
        Merge two YAML maps (CommentedMap-wrapped dicts).

        Parameters:
        1. lhs (CommentedMap) The merge target.
        2. rhs (CommentedMap) The merge source.
        3. path (YAMLPath) Location within the DOM where this merge is taking
           place.

        Returns:  (CommentedMap) The merged result.

        Raises:
        - `MergeException` when a clean merge is impossible.
        """
        if not isinstance(lhs, CommentedMap):
            raise MergeException(
                "Impossible to add Hash data to non-Hash destination.", path)

        self.logger.debug(
            "Merging INTO dict with keys: {}:".format(", ".join([
                    str(k.value) if isinstance(k, TaggedScalar)
                    else str(k)
                    for k in lhs.keys()])),
            data=lhs, prefix="Merger::_merge_dicts:  ",
            header="--------------------")
        self.logger.debug(
            "Merging FROM dict with keys: {}:".format(", ".join([
                    str(k.value) if isinstance(k, TaggedScalar)
                    else str(k)
                    for k in rhs.keys()])),
            data=rhs, prefix="Merger::_merge_dicts:  ",
            footer="====================")

        # Delete all internal YAML merge reference keys lest any later
        # .insert() operation on LHS inexplicably convert them from reference
        # to concrete keys.  This seems like a bug in ruamel.yaml...
        self._delete_mergeref_keys(lhs)

        # Assume deep merge until a node's merge rule indicates otherwise
        buffer: List[Tuple[Any, Any]] = []
        buffer_pos = 0
        for key, val in rhs.non_merged_items():
            path_next = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            if key in lhs:
                # Write the buffer if populated
                for b_key, b_val in buffer:
                    self.logger.debug(
                        "Merger::_merge_dicts:  Inserting key, {}, from"
                        " buffer to position, {}, at path, {}."
                        .format(b_key, buffer_pos, path_next),
                        header="INSERT " * 15)
                    self.logger.debug(
                        "Before INSERT, the LHS document was:",
                        data=lhs, prefix="Merger::_merge_dicts:  ")
                    self.logger.debug(
                        "... and before INSERT, the incoming value will be:",
                        data=b_val, prefix="Merger::_merge_dicts:  ")
                    lhs.insert(buffer_pos, b_key, b_val)
                    self.logger.debug(
                        "After INSERT, the LHS document became:",
                        data=lhs, prefix="Merger::_merge_dicts:  ")
                    buffer_pos += 1
                buffer = []

                # Short-circuit the deep merge if a different merge rule
                # applies to this node.
                node_coord = NodeCoords(val, rhs, key)
                merge_mode = (
                    self.config.hash_merge_mode(node_coord)
                    if isinstance(val, CommentedMap)
                    else self.config.set_merge_mode(node_coord)
                    if isinstance(val, CommentedSet)
                    else self.config.aoh_merge_mode(node_coord)
                )
                self.logger.debug("Merger::_merge_dicts:  Got merge mode, {}."
                                  .format(merge_mode))
                if merge_mode in (
                    HashMergeOpts.LEFT, AoHMergeOpts.LEFT, SetMergeOpts.LEFT
                ):
                    continue
                if merge_mode in (
                    HashMergeOpts.RIGHT, AoHMergeOpts.RIGHT, SetMergeOpts.RIGHT
                ):
                    self.logger.debug(
                        "Merger::_merge_dicts:  Overwriting key, {}, at path,"
                        " {}.".format(key, path_next),
                        header="OVERWRITE " * 15)
                    lhs[key] = val
                    continue

                if isinstance(val, CommentedMap):
                    lhs[key] = self._merge_dicts(lhs[key], val, path_next)

                    # Synchronize any YAML Tag
                    self.logger.debug(
                        "Merger::_merge_dicts:  Setting LHS tag from {} to {}."
                        .format(lhs[key].tag.value, val.tag.value))
                    lhs[key].yaml_set_tag(val.tag.value)

                    self.logger.debug(
                        "Document BEFORE calling combine_merge_anchors:",
                        data=lhs, prefix="Merger::_merge_dicts:  ",
                        header="+------------------+")
                    Anchors.combine_merge_anchors(lhs[key], val)
                    self.logger.debug(
                        "Document AFTER calling combine_merge_anchors:",
                        data=lhs, prefix="Merger::_merge_dicts:  ",
                        footer="+==================+")
                elif isinstance(val, CommentedSeq):
                    lhs[key] = self._merge_lists(
                        lhs[key], val, path_next, parent=rhs, parentref=key)

                    # Synchronize any YAML Tag
                    self.logger.debug(
                        "Merger::_merge_dicts:  Setting LHS tag from {} to {}."
                        .format(lhs[key].tag.value, val.tag.value))
                    lhs[key].yaml_set_tag(val.tag.value)
                elif isinstance(val, CommentedSet):
                    lhs[key] = self._merge_sets(
                        lhs[key], val, path_next, node_coord)

                    # Synchronize any YAML Tag
                    self.logger.debug(
                        "Merger::_merge_dicts:  Setting LHS tag from {} to {}."
                        .format(lhs[key].tag.value, val.tag.value))
                    lhs[key].yaml_set_tag(val.tag.value)
                else:
                    self.logger.debug(
                        "Merger::_merge_dicts:  Updating key, {}, at path,"
                        " {}.".format(key, path_next), header="UPDATE " * 15)
                    self.logger.debug(
                        "Before UPDATE, the LHS document was:",
                        data=lhs, prefix="Merger::_merge_dicts:  ")
                    self.logger.debug(
                        "... and before UPDATE, the incoming value will be:",
                        data=val, prefix="Merger::_merge_dicts:  ")
                    lhs[key] = val
                    self.logger.debug(
                        "After UPDATE, the LHS document became:",
                        data=lhs, prefix="Merger::_merge_dicts:  ")
            else:
                # LHS lacks the RHS key.  Buffer this key-value pair in order
                # to insert it ahead of whatever key(s) follow this one in RHS
                # to keep anchor definitions before their aliases.
                buffer.append((key, val))

            buffer_pos += 1

        # Write any remaining buffered content to the end of LHS
        for b_key, b_val in buffer:
            self.logger.debug(
                "Merger::_merge_dicts:  Appending key, {}, from buffer at"
                " path, {}.".format(b_key, path), header="APPEND " * 15)
            lhs[b_key] = b_val

        self.logger.debug(
            "Completed merge result for path, {}:".format(path),
            data=lhs, prefix="Merger::_merge_dicts:  ")

        return lhs

    def _merge_simple_lists(
        self, lhs: CommentedSeq, rhs: CommentedSeq, path: YAMLPath,
        node_coord: NodeCoords
    ) -> CommentedSeq:
        """
        Merge two CommentedSeq-wrapped lists of Scalars or CommentedSeqs.

        Parameters:
        1. lhs (CommentedSeq) The merge target.
        2. rhs (CommentedSeq) The merge source.
        3. path (YAMLPath) Location within the DOM where this merge is taking
           place.
        4. node_coord (NodeCoords) The RHS root node, its parent, and reference
           within its parent; used for config lookups.

        Returns: (CommentedSeq) The merged result.

        Raises:
        - `MergeException` when a clean merge is impossible.
        """
        if not isinstance(lhs, CommentedSeq):
            raise MergeException(
                "Impossible to add Array data to non-Array destination.", path)

        merge_mode = self.config.array_merge_mode(node_coord)
        if merge_mode is ArrayMergeOpts.LEFT:
            return lhs
        if merge_mode is ArrayMergeOpts.RIGHT:
            return rhs

        tagless_lhs = Nodes.tagless_elements(lhs)
        for idx, ele in enumerate(rhs):
            path_next = path + "[{}]".format(idx)
            self.logger.debug(
                "Processing element {} at {}.".format(idx, path_next),
                prefix="Merger::_merge_simple_lists:  ", data=ele)

            if merge_mode is ArrayMergeOpts.UNIQUE:
                cmp_val = ele
                if isinstance(ele, TaggedScalar):
                    cmp_val = ele.value

                self.logger.debug(
                    "Looking for comparison value, {}, in:".format(cmp_val),
                    prefix="Merger::_merge_simple_lists:  ", data=tagless_lhs)

                if cmp_val in tagless_lhs:
                    lhs = CommentedSeq([ele
                           if (e == cmp_val
                               or (isinstance(e, TaggedScalar)
                                   and e.value == cmp_val)
                           ) else e
                           for e in lhs])
                else:
                    lhs.append(ele)
                continue
            lhs.append(ele)
        return lhs

    # pylint: disable=locally-disabled,too-many-branches
    def _merge_arrays_of_hashes(
        self, lhs: CommentedSeq, rhs: CommentedSeq, path: YAMLPath,
        node_coord: NodeCoords
    ) -> CommentedSeq:
        """
        Merge two Arrays-of-Hashes.

        This is a deep merge operation.  Each dict is treated as a record with
        an identity key.  RHS records are merged with LHS records for which the
        identity key matches.  As such, an identity key is required in both LHS
        and RHS records.  This key is configurable.  When there is no LHS match
        for an RHS key, the RHS record is appended to the LHS list.

        Parameters:
        1. lhs (CommentedSeq) The merge target.
        2. rhs (CommentedSeq) The merge source.
        3. path (YAMLPath) Location within the DOM where this merge is taking
           place.
        4. node_coord (NodeCoords) The RHS root node, its parent, and reference
           within its parent; used for config lookups.

        Returns:  (CommentedSeq) The merged result.

        Raises:
        - `MergeException` when a clean merge is impossible.
        """
        if not isinstance(lhs, CommentedSeq):
            raise MergeException(
                "Impossible to add Array-of-Hash data to non-Array"
                " destination."
                , path)

        self.logger.debug(
            "Merging {} Hash(es) at {}.".format(len(rhs), path),
            prefix="Merger::_merge_arrays_of_hashes:  ", data=rhs)

        id_key: str = ""
        if len(rhs) > 0 and isinstance(rhs[0], CommentedMap):
            id_key = self.config.aoh_merge_key(
                NodeCoords(rhs[0], rhs, 0), rhs[0])
            self.logger.debug(
                "Merger::_merge_arrays_of_hashes:  RHS AoH yielded id_key:"
                "  {}.".format(id_key))

        merge_mode = self.config.aoh_merge_mode(node_coord)
        for idx, ele in enumerate(rhs):
            path_next = path + "[{}]".format(idx)
            self.logger.debug(
                "Processing element #{} at {}.".format(idx, path_next),
                prefix="Merger::_merge_arrays_of_hashes:  ", data=ele)

            if merge_mode is AoHMergeOpts.DEEP:
                if id_key in ele:
                    id_val = Nodes.tagless_value(ele[id_key])
                else:
                    raise MergeException(
                        "Mandatory identity key, {}, not present in Hash with"
                        " keys:  {}."
                        .format(id_key, ", ".join(ele.keys()))
                        , path_next
                    )

                merged_hash = False
                for lhs_hash in (
                    lhs_hash for lhs_hash in lhs
                    if isinstance(lhs_hash, CommentedMap)
                    and id_key in lhs_hash
                    and Nodes.tagless_value(lhs_hash[id_key]) == id_val
                ):
                    self._merge_dicts(lhs_hash, ele, path_next)
                    merged_hash = True

                    # Synchronize YAML Tags
                    lhs_hash.yaml_set_tag(ele.tag.value)
                    break
                if not merged_hash:
                    Nodes.append_list_element(lhs, ele,
                        ele.anchor.value if hasattr(ele, "anchor") else None)
            elif merge_mode is AoHMergeOpts.UNIQUE:
                if ele not in lhs:
                    Nodes.append_list_element(
                        lhs, ele,
                        ele.anchor.value if hasattr(ele, "anchor") else None)
            else:
                Nodes.append_list_element(lhs, ele,
                    ele.anchor.value if hasattr(ele, "anchor") else None)
        return lhs

    def _merge_lists(
        self, lhs: CommentedSeq, rhs: CommentedSeq, path: YAMLPath,
        **kwargs: Any
    ) -> CommentedSeq:
        """
        Merge two lists; understands lists-of-dicts.

        Parameters:
        1. lhs (CommentedSeq) The list to merge into.
        2. rhs (CommentedSeq) The list to merge from.
        3. path (YAMLPath) Location of the `rhs` source list within its DOM.

        Keyword Arguments:
        * parent (Any) Parent node of `rhs`
        * parentref (Any) Child Key or Index of `rhs` within `parent`.

        Returns:  (CommentedSeq) The merged result.
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        node_coord = NodeCoords(rhs, parent, parentref)
        if len(rhs) > 0:
            if isinstance(rhs[0], CommentedMap):
                # This list is an Array-of-Hashes
                return self._merge_arrays_of_hashes(lhs, rhs, path, node_coord)

            # This list is an Array-of-Arrays or a simple list of Scalars
            return self._merge_simple_lists(lhs, rhs, path, node_coord)

        # No RHS list
        return lhs

    def _merge_sets(
        self, lhs: CommentedSet, rhs: CommentedSet, path: YAMLPath,
        node_coord: NodeCoords
    ) -> CommentedSet:
        """
        Merge two YAML sets (CommentedSet-wrapped sets).

        Parameters:
        1. lhs (CommentedSet) The merge target.
        2. rhs (CommentedSet) The merge source.
        3. path (YAMLPath) Location within the DOM where this merge is taking
           place.
        4. node_coord (NodeCoords) The RHS root node, its parent, and reference
           within its parent; used for config lookups.

        Returns:  (CommentedSet) The merged result.

        Raises:
        - `MergeException` when a clean merge is impossible.
        """
        merge_mode = self.config.set_merge_mode(node_coord)
        if merge_mode is SetMergeOpts.LEFT:
            return lhs
        if merge_mode is SetMergeOpts.RIGHT:
            return rhs

        tagless_lhs = Nodes.tagless_elements(list(lhs))
        for ele in rhs:
            path_next = (path +
                YAMLPath.escape_path_section(ele, path.seperator))
            self.logger.debug(
                "Processing set element {} at {}.".format(ele, path_next),
                prefix="Merger::_merge_sets:  ", data=ele)

            cmp_val = ele
            if isinstance(ele, TaggedScalar):
                cmp_val = ele.value

            self.logger.debug(
                "Looking for comparison value, {}, in:".format(cmp_val),
                prefix="Merger::_merge_sets:  ", data=tagless_lhs)

            if cmp_val in tagless_lhs:
                continue
            lhs.add(ele)
        return lhs

    def _calc_unique_anchor(self, anchor: str, known_anchors: Set[str]) -> str:
        """
        Generate a unique anchor name within a document pair.

        Parameters:
        1. anchor (str) The original anchor name.
        2. known_anchors (dict) Dictionary of every anchor already in the
           document.

        Returns:  (str) The new, unique anchor name.
        """
        self.logger.debug(
            "Preexisting Anchors:", prefix="Merger::_calc_unique_anchor:  ",
            data=known_anchors)
        aid = 1
        while anchor in known_anchors:
            anchor = "{}_{}".format(anchor, aid)
            aid += 1
        return anchor

    def _resolve_anchor_conflicts(self, rhs: Any) -> None:
        """
        Resolve anchor conflicts between this and another document.

        Parameters:
        1. rhs (Any) The other document to consolidate with this one.

        Returns:  N/A
        """
        lhs_anchors: Dict[str, Any] = {}
        Anchors.scan_for_anchors(self.data, lhs_anchors)
        self.logger.debug(
            "LHS Anchors:", prefix="Merger::_resolve_anchor_conflicts:  ",
            data=lhs_anchors)

        rhs_anchors: Dict[str, Any] = {}
        Anchors.scan_for_anchors(rhs, rhs_anchors)
        self.logger.debug(
            "RHS Anchors:", prefix="Merger::_resolve_anchor_conflicts:  ",
            data=rhs_anchors)

        for anchor in [anchor
                for anchor in rhs_anchors
                if anchor in lhs_anchors
        ]:
            # It is only a conflict if the value differs; however, the
            # value may be a scalar, list, or dict.  Further, lists and
            # dicts may contain other aliased values which must also be
            # checked for equality (or pointing at identical anchors).
            lhs_anchor = lhs_anchors[anchor]
            rhs_anchor = rhs_anchors[anchor]
            conflict_mode = self.config.anchor_merge_mode()

            self.logger.debug(
                "Anchor is in both documents:",
                prefix="Merger::_resolve_anchor_conflicts:  ", data=anchor)
            self.logger.debug(
                "lhs_anchor:", prefix="Merger::_resolve_anchor_conflicts:  ",
                data=lhs_anchor)
            self.logger.debug(
                "rhs_anchor:", prefix="Merger::_resolve_anchor_conflicts:  ",
                data=rhs_anchor)

            anchors_match = True
            lhs_is_tagged = isinstance(lhs_anchor, TaggedScalar)
            rhs_is_tagged = isinstance(rhs_anchor, TaggedScalar)
            if lhs_is_tagged != rhs_is_tagged:
                anchors_match = False
            elif lhs_is_tagged:
                anchors_match = (
                    (lhs_anchor.value == rhs_anchor.value)
                    and (lhs_anchor.tag.value == rhs_anchor.tag.value))
            else:
                anchors_match = lhs_anchor == rhs_anchor

            if not anchors_match:
                if conflict_mode is AnchorConflictResolutions.RENAME:
                    self.logger.debug(
                        "Anchor {} conflict; will RENAME anchors."
                        .format(anchor),
                        prefix="Merger::_resolve_anchor_conflicts:  ")
                    Anchors.rename_anchor(
                        rhs, anchor,
                        self._calc_unique_anchor(
                            anchor,
                            set(lhs_anchors.keys())
                            .union(set(rhs_anchors.keys()))
                        )
                    )
                elif conflict_mode is AnchorConflictResolutions.LEFT:
                    self.logger.debug(
                        "Anchor {} conflict; LEFT will override."
                        .format(anchor),
                        prefix="Merger::_resolve_anchor_conflicts:  ")
                    Anchors.replace_anchor(rhs, rhs_anchor, lhs_anchor)
                elif conflict_mode is AnchorConflictResolutions.RIGHT:
                    self.logger.debug(
                        "Anchor {} conflict; RIGHT will override."
                        .format(anchor),
                        prefix="Merger::_resolve_anchor_conflicts:  ")
                    Anchors.replace_anchor(self.data, lhs_anchor, rhs_anchor)
                else:
                    raise MergeException(
                        "Aborting due to anchor conflict with, {}."
                        .format(anchor))
            else:
                self.logger.debug(
                    "Merger::_resolve_anchor_conflicts:  Anchor {} is"
                    " symmetric; RIGHT will override to eliminate spurious"
                    " anchor re-definition.".format(anchor))
                # While the anchors are identical, the reference nodes are not.
                # So, overwrite all matching LHS nodes with their RHS
                # equivalents in order to stave off spurious anchor
                # re-definitions.
                Anchors.replace_anchor(self.data, lhs_anchor, rhs_anchor)

    def _insert_dict(
        self, insert_at: YAMLPath,
        lhs: Union[CommentedMap, CommentedSeq, CommentedSet],
        rhs: CommentedMap
    ) -> bool:
        """Insert an RHS dict merge result into the LHS document."""
        merge_performed = False
        merged_data: Union[
            CommentedMap, CommentedSeq, CommentedSet
        ] = CommentedMap()
        if isinstance(lhs, CommentedSeq):
            # Merge a dict into a list
            self.logger.debug(
                "Merger::_insert_dict:  Merging a dict into a list.")
            merged_data = self._merge_lists(
                lhs, CommentedSeq([rhs]), insert_at)
            merge_performed = True
        elif isinstance(lhs, CommentedSet):
            # Merge a dict into a set; this is destructive
            raise MergeException(
                "Merging a Hash into a Set is destructive to the"
                " source Hash because only the keys would be"
                " preserved.  Please adjust your merge to target a"
                " suitable node.", insert_at)
        else:
            # Merge a dict into a dict
            self.logger.debug(
                "Merger::_insert_dict:  Merging a dict into a dict.")

            merge_mode = self.config.hash_merge_mode(
                NodeCoords(rhs, None, None))
            if merge_mode is HashMergeOpts.LEFT:
                self.logger.debug(
                    "Configured mode short-circuits the merge; returning LHS:"
                    , prefix="Merger::_insert_dict:  "
                    , data=lhs)
                merged_data = lhs
            elif merge_mode is HashMergeOpts.RIGHT:
                self.logger.debug(
                    "Configured mode short-circuits the merge; returning RHS:"
                    , prefix="Merger::_insert_dict:  "
                    , data=rhs)
                merged_data = rhs
            else:
                merged_data = self._merge_dicts(lhs, rhs, insert_at)

            merge_performed = True

        # Synchronize YAML Tags
        self.logger.debug(
            "Merger::_insert_dict:  Setting LHS tag from {} to {}."
            .format(lhs.tag.value, rhs.tag.value))
        lhs.yaml_set_tag(rhs.tag.value)

        if insert_at.is_root:
            self.data = merged_data
        return merge_performed

    def _insert_list(
        self, insert_at: YAMLPath,
        lhs: Union[CommentedMap, CommentedSeq, CommentedSet],
        rhs: CommentedSeq
    ) -> bool:
        """Insert an RHS list merge result into the LHS document."""
        merge_performed = False
        merged_data: Union[CommentedSeq, CommentedSet, None] = None
        if isinstance(lhs, CommentedSeq):
            # Merge list into list
            self.logger.debug(
                "Merger::_insert_list:  Merging a list into a list.")
            merged_data = self._merge_lists(
                lhs, rhs, insert_at)
            merge_performed = True
        elif isinstance(lhs, CommentedSet):
            # Merge list into set
            self.logger.debug(
                "Merger::_insert_list:  Merging a list into a set.")
            mset = CommentedSet()
            for ele in rhs:
                mset.add(ele)
            merged_data = self._merge_sets(
                lhs, mset, insert_at, NodeCoords(rhs, None, None))
            merge_performed = True
        else:
            # Merge list into hash
            raise MergeException(
                "Impossible to merge an Array into a Hash without a"
                " key given to receive the new elements.", insert_at)

        # Synchronize any YAML Tag
        self.logger.debug(
            "Merger::_insert_list:  Setting LHS tag from {} to {}."
            .format(lhs.tag.value, rhs.tag.value))
        lhs.yaml_set_tag(rhs.tag.value)

        if insert_at.is_root:
            self.data = merged_data
        return merge_performed

    def _insert_set(
        self, insert_at: YAMLPath,
        lhs: Union[CommentedMap, CommentedSeq, CommentedSet],
        rhs: CommentedSet
    ) -> bool:
        """Insert an RHS list merge result into the LHS document."""
        merge_performed = False
        merged_data: Union[
            CommentedSeq, CommentedMap, CommentedSet, None] = None
        if isinstance(lhs, CommentedSeq):
            self.logger.debug(
                "Merger::_insert_set:  Merging a set into a list.")
            merge_list = []
            for ele in rhs:
                merge_list.append(ele)
            merged_data = self._merge_lists(
                lhs, CommentedSeq(merge_list), insert_at)
            merge_performed = True
        elif isinstance(lhs, CommentedMap):
            self.logger.debug(
                "Merger::_insert_set:  Merging a set into a dict.")
            merge_dict: Dict[str, None] = {}
            for ele in rhs:
                merge_dict[ele] = None
            merged_data = self._merge_dicts(
                lhs, CommentedMap(merge_dict), insert_at)
            merge_performed = True
        else:
            self.logger.debug(
                "Merger::_insert_set:  Merging a set into a set.")
            merged_data = self._merge_sets(lhs, rhs, insert_at, NodeCoords(
                rhs, None, None
            ))
            merge_performed = True

        # Synchronize any YAML Tag
        self.logger.debug(
            "Merger::_insert_set:  Setting LHS tag from {} to {}."
            .format(lhs.tag.value, rhs.tag.value))
        lhs.yaml_set_tag(rhs.tag.value)

        if insert_at.is_root:
            self.data = merged_data
        return merge_performed

    def _insert_scalar(
        self, insert_at: YAMLPath, lhs: Any, lhs_proc: Processor, rhs: Any
    ) -> bool:
        """Insert an RHS scalar into the LHS document."""
        merge_performed = False
        if isinstance(lhs, CommentedSeq):
            self.logger.debug(
                "Merger::_insert_scalar:  Merging a scalar into a list.")
            Nodes.append_list_element(lhs, rhs)
            merge_performed = True
        elif isinstance(lhs, CommentedSet):
            self.logger.debug(
                "Merger::_insert_scalar:  Merging a scalar into a set.")
            self._merge_sets(
                lhs, CommentedSet([rhs]), insert_at,
                NodeCoords(rhs, None, None))
            merge_performed = True
        elif isinstance(lhs, CommentedMap):
            ex_message = (
                "Impossible to add Scalar value, {}, to a Hash without"
                " a key.  Change the value to a 'key: value' pair, a"
                " '{{key: value}}' Hash, or change the merge target to"
                " an Array or other Scalar value."
                ).format(rhs)
            if len(str(rhs)) < 1 and not sys.stdin.isatty():
                ex_message += (
                    "  You may be seeing this because your workflow"
                    " inadvertently opened a STDIN handle to {}.  If"
                    " this may be the case, try adding --nostdin or -S"
                    " so as to block unintentional STDIN reading."
                ).format(basename(sys.argv[0]))
            raise MergeException(ex_message, insert_at)
        else:
            lhs_proc.set_value(insert_at, rhs)
            merge_performed = True
        return merge_performed

    def _get_merge_target_nodes(
        self, insert_at: YAMLPath, lhs_proc: Processor, rhs: Any
    ) -> List[NodeCoords]:
        """Get a list of LHS insertion points for merge results."""
        # Loop through all insertion points and the elements in RHS
        nodes: List[NodeCoords] = []
        for node_coord in lhs_proc.get_nodes(
            insert_at, default_value=rhs
        ):
            nodes.append(node_coord)

        self.logger.debug(
            "Targetting these destinations for merge results:",
            prefix="Merger::_get_merge_target_nodes:  ",
            data=nodes)
        return nodes

    def merge_with(self, rhs: Any) -> None:
        """
        Merge this document with another.

        Parameters:
        1. rhs (Any) The document to merge into this one.

        Returns:  N/A

        Raises:
        - `MergeException` when a clean merge is impossible.
        """
        # Do nothing when RHS is None (empty document)
        if rhs is None:
            return

        # Remove all comments (no sensible way to merge them)
        Parsers.delete_all_comments(rhs)

        # When LHS is None (empty document), just dump all of RHS into it,
        # honoring any --mergeat|-m location as best as possible.
        insert_at = self.config.get_insertion_point()
        if self.data is None:
            self.logger.debug(
                "Replacing None data with:", prefix="Merger::merge_with:  ",
                data=rhs, data_header="     *****")
            self.data = Nodes.build_next_node(insert_at, 0, rhs)
            self.logger.debug(
                "Merged document is now:", prefix="Merger::merge_with:  ",
                data=self.data, footer="     ***** ***** *****")
            if isinstance(rhs, (dict, list, CommentedSet, set)):
                # Only Scalar values need further processing
                return

        # Resolve any anchor conflicts
        self._resolve_anchor_conflicts(rhs)

        # Prepare the merge rules
        self.config.prepare(rhs)

        # Merge into each insertion point
        merge_performed = False
        lhs_proc = Processor(self.logger, self.data)
        for node_coord in self._get_merge_target_nodes(
            insert_at, lhs_proc, rhs
        ):
            target_node = node_coord.node
            Parsers.set_flow_style(
                rhs, (target_node.fa.flow_style()
                      if hasattr(target_node, "fa")
                      else None))

            if isinstance(rhs, CommentedMap):
                # The RHS document root is a dict
                merge_performed = self._insert_dict(
                    insert_at, target_node, rhs)
            elif isinstance(rhs, CommentedSeq):
                # The RHS document root is a list
                merge_performed = self._insert_list(
                    insert_at, target_node, rhs)
            elif isinstance(rhs, CommentedSet):
                # The RHS document is a set
                merge_performed = self._insert_set(
                    insert_at, target_node, rhs)
            else:
                # The RHS document root is a Scalar value
                merge_performed = self._insert_scalar(
                    insert_at, target_node, lhs_proc, rhs)

        self.logger.debug(
            "Completed merge operation, resulting in document:",
            prefix="Merger::merge_with:  ", data=self.data)

        if not merge_performed:
            raise MergeException(
                "A merge was not performed.  Ensure your target path matches"
                " at least one node in the left document(s).", insert_at)

    def prepare_for_dump(
        self, yaml_writer: Any, output_file: str = ""
    ) -> OutputDocTypes:
        """
        Prepare this merged document and its writer for final rendering.

        This coalesces the YAML writer's settings to, in particular,
        distinguish between YAML and JSON.  It will also force demarcation of
        every String key and value within the document when the output will be
        JSON.

        Parameters:
        1. yaml_writer (ruamel.yaml.YAML) The YAML document writer

        Returns:  (OutputDocTypes) One of:
          * OutputDocTypes.JSON:  The document and yaml_writer are JSON format.
          * OutputDocTypes.YAML:  The document and yaml_writer are YAML format.
        """
        # Check whether the user is forcing an output format
        doc_format = self.config.get_document_format()
        if doc_format is OutputDocTypes.AUTO:
            # Identify by file-extension, if it indicates a known type
            file_extension = (Path(output_file).suffix.lower()
                              if output_file else "")
            if file_extension in [".json", ".yaml", ".yml"]:
                is_flow = file_extension == ".json"
            else:
                # Check whether the document root is in flow or block format
                is_flow = True
                if hasattr(self.data, "fa"):
                    is_flow = self.data.fa.flow_style()
        else:
            is_flow = doc_format is OutputDocTypes.JSON

        if is_flow:
            # Dump the document as true JSON and reload it; this automatically
            # exlodes all aliases.
            xfer_buffer = StringIO()
            json.dump(Parsers.jsonify_yaml_data(self.data), xfer_buffer)
            xfer_buffer.seek(0)
            self.data = yaml_writer.load(xfer_buffer)

            # Ensure the writer doesn't emit a YAML Start-of-Document marker
            yaml_writer.explicit_start = False
        else:
            # Ensure block style output
            Parsers.set_flow_style(self.data, False)

            # When writing YAML, ensure the document start mark is emitted
            yaml_writer.explicit_start = True

        return OutputDocTypes.JSON if is_flow else OutputDocTypes.YAML

    @classmethod
    def set_flow_style(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Parsers.set_flow_style(*args)

    @classmethod
    def delete_all_comments(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Parsers.delete_all_comments(*args)

    @classmethod
    def combine_merge_anchors(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Anchors.combine_merge_anchors(*args)

    @classmethod
    def rename_anchor(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Anchors.rename_anchor(*args)

    @classmethod
    def replace_anchor(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Anchors.replace_anchor(*args)

    @classmethod
    def scan_for_anchors(cls, *args):
        """Relay function call to static method."""
        if not cls.depwarn_printed:
            cls.depwarn_printed = True
            print(Merger.DEPRECATION_WARNING, file=sys.stderr)
        Anchors.scan_for_anchors(*args)



import pytest

from yamlpath.merger.enums.aohmergeopts import (
	AoHMergeOpts)


class Test_merger_enums_aohmergeopts():
	"""Tests for the AoHMergeOpts enumeration."""

	def test_get_names(self):
		assert AoHMergeOpts.get_names() == [
			"ALL",
			"DEEP",
			"LEFT",
			"RIGHT",
			"UNIQUE",
		]

	def test_get_choices(self):
		assert AoHMergeOpts.get_choices() == [
			"all",
			"deep",
			"left",
			"right",
			"unique",
		]

	@pytest.mark.parametrize("input,output", [
		("ALL", AoHMergeOpts.ALL),
		("DEEP", AoHMergeOpts.DEEP),
		("LEFT", AoHMergeOpts.LEFT),
		("RIGHT", AoHMergeOpts.RIGHT),
		("UNIQUE", AoHMergeOpts.UNIQUE),
	])
	def test_from_str(self, input, output):
		assert output == AoHMergeOpts.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			AoHMergeOpts.from_str("NO SUCH NAME")



"""
Implements DifferConfig.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
import configparser
from typing import Any, Dict, Tuple, Union
from argparse import Namespace

from yamlpath.exceptions import YAMLPathException
from yamlpath.differ.enums import AoHDiffOpts, ArrayDiffOpts
from yamlpath import Processor, YAMLPath
from yamlpath.wrappers import ConsolePrinter, NodeCoords


class DifferConfig:
    """Config file processor for the Differ."""

    def __init__(self, logger: ConsolePrinter, args: Namespace) -> None:
        """
        Instantiate this class into an object.

        Parameters:
        1. logger (ConsolePrinter) Instance of ConsoleWriter or subclass
        2. args (dict) Default options for diff rules

        Returns:  N/A
        """
        self.log = logger
        self.args = args
        self.config: Union[None, configparser.ConfigParser] = None
        self.rules: Dict[NodeCoords, str] = {}
        self.keys: Dict[NodeCoords, str] = {}

        self._load_config()

    def array_diff_mode(self, node_coord: NodeCoords) -> ArrayDiffOpts:
        """
        Get Array diff mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (ArrayDiffOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        diff_rule = self._get_rule_for(node_coord)
        if diff_rule:
            self.log.debug(
                "DifferConfig::array_diff_mode:  Matched {}"
                .format(diff_rule))
            return ArrayDiffOpts.from_str(diff_rule)
        self.log.debug("DifferConfig::array_diff_mode:  NOT Matched")
        if hasattr(self.args, "arrays") and self.args.arrays:
            return ArrayDiffOpts.from_str(self.args.arrays)
        if (self.config is not None
                and "defaults" in self.config
                and "arrays" in self.config["defaults"]):
            return ArrayDiffOpts.from_str(self.config["defaults"]["arrays"])
        return ArrayDiffOpts.POSITION

    def aoh_diff_mode(self, node_coord: NodeCoords) -> AoHDiffOpts:
        """
        Get Array-of-Hashes diff mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (AoHDiffOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        diff_rule = self._get_rule_for(node_coord)
        if diff_rule:
            self.log.debug(
                "DifferConfig::aoh_diff_mode:  Matched {}"
                .format(diff_rule))
            return AoHDiffOpts.from_str(diff_rule)
        self.log.debug("DifferConfig::aoh_diff_mode:  NOT Matched")
        if hasattr(self.args, "aoh") and self.args.aoh:
            return AoHDiffOpts.from_str(self.args.aoh)
        if (self.config is not None
                and "defaults" in self.config
                and "aoh" in self.config["defaults"]):
            return AoHDiffOpts.from_str(self.config["defaults"]["aoh"])
        return AoHDiffOpts.POSITION

    def aoh_diff_key(self, node_coord: NodeCoords) -> Tuple[str, bool]:
        """
        Get the user-defined identity key for Array-of-Hashes comparisons.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns: (str, bool) The identity key field name and whether the key
          is specifically user-defined or inferred from the data.
        """
        # Check the user config for a specific key; fallback to first key.
        is_user_key = True
        diff_key = self._get_key_for(node_coord)
        if not diff_key:
            # This node may be a child of one of the registered keys.  That
            # registered key's node will match this node's parent.
            for eval_nc, eval_key in self.keys.items():
                if node_coord.parent == eval_nc.node:
                    diff_key = eval_key
                    break

        node = node_coord.node
        if not diff_key and isinstance(node, dict) and len(node.keys()) > 0:
            # Fallback to using the first key of the dict as an identity key
            is_user_key = False
            diff_key = list(node)[0]

        return (diff_key, is_user_key)

    def prepare(self, data: Any) -> None:
        """
        Load references to all nodes which match config rules.

        Parameters:
        1. data (Any) The DOM for which to load configuration.

        Returns:  N/A
        """
        if self.config is None:
            return

        # Eliminate previous rules and keys to limit scanning to only those
        # nodes which exist within this new document.
        self.rules = {}
        self.keys = {}

        # Load new rules and keys
        proc = Processor(self.log, data)
        self._prepare_user_rules(proc, "rules", self.rules)
        self._prepare_user_rules(proc, "keys", self.keys)

    def _prepare_user_rules(
        self, proc: Processor, section: str, collector: dict
    ) -> None:
        """
        Identify DOM nodes matching user-defined diff rules.

        Parameters:
        1. proc (Processor) Reference to the DOM Processor.
        2. section (str) User-configuration file section defining the diff
           rules to apply.
        3. collector (dict) Storage collector for matching nodes.

        Returns:  N/A
        """
        if self.config is None or not section in self.config:
            self.log.warning(
                "User-specified configuration file has no {} section."
                .format(section))
            return

        for rule_key in self.config[section]:
            rule_value = self.config[section][rule_key]

            if "=" in rule_value:
                # There were at least two = signs on the configuration line
                conf_line = rule_key + "=" + rule_value
                delim_pos = conf_line.rfind("=")
                rule_key = conf_line[0:delim_pos].strip()
                rule_value = conf_line[delim_pos + 1:].strip()
                self.log.debug(
                    "DifferConfig::_prepare_user_rules:  Reconstituted"
                    " configuration line '{}' to extract adjusted key '{}'"
                    " with value '{}'".format(conf_line, rule_key, rule_value))

            rule_path = YAMLPath(rule_key)
            yaml_path = YAMLPath(rule_path)
            self.log.debug(
                "DifferConfig::_prepare_user_rules:  Matching '{}' nodes to"
                " YAML Path '{}' from key, {}."
                .format(section, yaml_path, rule_key))
            try:
                for node_coord in proc.get_nodes(yaml_path, mustexist=True):
                    self.log.debug(
                        "Node will have comparisons rule, {}:"
                        .format(rule_value),
                        prefix="DifferConfig::_prepare_user_rules:  ",
                        data=node_coord.node)
                    collector[node_coord] = rule_value

            except YAMLPathException:
                self.log.warning("{} YAML Path matches no nodes:  {}"
                                .format(section, yaml_path))

        self.log.debug(
            "Matched rules to nodes:",
            prefix="DifferConfig::_prepare_user_rules:  ")
        for node_coord, diff_rule in collector.items():
            self.log.debug(
                "... RULE:  {}".format(diff_rule),
                prefix="DifferConfig::_prepare_user_rules:  ")
            self.log.debug(
                "... NODE:", data=node_coord,
                prefix="DifferConfig::_prepare_user_rules:  ")

    def _load_config(self) -> None:
        """Load the external configuration file."""
        config = configparser.ConfigParser()

        # Load the configuration file when one is specified
        config_file = (
            self.args.config
            if hasattr(self.args, "config")
            else None)

        if config_file:
            config.read(config_file)
            if config.sections():
                self.config = config

    def _get_config_for(self, node_coord: NodeCoords, section: dict) -> str:
        """
        Get user configuration applicable to a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.
        2. section (dict) The configuration section to query.

        Returns: (str) The requested configuration.
        """
        if self.config is None:
            return ""

        for rule_coord, rule_config in section.items():
            if rule_coord.node == node_coord.node \
                    and rule_coord.parent == node_coord.parent \
                    and rule_coord.parentref == node_coord.parentref:
                return str(rule_config)

        return ""

    def _get_rule_for(self, node_coord: NodeCoords) -> str:
        """
        Get a user configured diff rule for a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.

        Returns: (str) The requested configuration.
        """
        self.log.debug(
            "Seeking rule for node:", prefix="DifferConfig::_get_rule_for:  ",
            header=" ")
        self.log.debug(
            "... NODE:", prefix="DifferConfig::_get_rule_for:  ",
            data=node_coord)
        return self._get_config_for(node_coord, self.rules)

    def _get_key_for(self, node_coord: NodeCoords) -> str:
        """
        Get a user configured diff identity key (field) for a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.

        Returns: (str) The requested configuration.
        """
        return self._get_config_for(node_coord, self.keys)



import pytest

import ruamel.yaml as ry

from yamlpath.enums import AnchorMatches, PathSearchMethods
from yamlpath.path import SearchTerms
from yamlpath.common import Searches

class Test_common_searches():
    """Tests for the Searches helper class."""

    ###
    # search_matches
    ###
    @pytest.mark.parametrize("match, method, needle, haystack", [
        (True, PathSearchMethods.CONTAINS, "a", "parents"),
        (True, PathSearchMethods.ENDS_WITH, "ts", "parents"),
        (True, PathSearchMethods.EQUALS, "parents", "parents"),
        (True, PathSearchMethods.EQUALS, 42, 42),
        (True, PathSearchMethods.EQUALS, "42", 42),
        (True, PathSearchMethods.EQUALS, 3.14159265385, 3.14159265385),
        (True, PathSearchMethods.EQUALS, "3.14159265385", 3.14159265385),
        (True, PathSearchMethods.EQUALS, True, True),
        (True, PathSearchMethods.EQUALS, "True", True),
        (True, PathSearchMethods.EQUALS, "true", True),
        (True, PathSearchMethods.EQUALS, False, False),
        (True, PathSearchMethods.EQUALS, "False", False),
        (True, PathSearchMethods.EQUALS, "false", False),
        (True, PathSearchMethods.GREATER_THAN, 2, 4),
        (True, PathSearchMethods.GREATER_THAN, "2", 4),
        (True, PathSearchMethods.GREATER_THAN, 2, "4"),
        (True, PathSearchMethods.GREATER_THAN, "2", "4"),
        (True, PathSearchMethods.GREATER_THAN, 2.1, 2.2),
        (True, PathSearchMethods.GREATER_THAN, "2.1", 2.2),
        (True, PathSearchMethods.GREATER_THAN, 2.1, "2.2"),
        (True, PathSearchMethods.GREATER_THAN, "2.1", "2.2"),
        (True, PathSearchMethods.GREATER_THAN, 2, 2.1),
        (True, PathSearchMethods.GREATER_THAN, "2", 2.1),
        (True, PathSearchMethods.GREATER_THAN, 2, "2.1"),
        (True, PathSearchMethods.GREATER_THAN, "2", "2.1"),
        (True, PathSearchMethods.GREATER_THAN, 2.9, 3),
        (True, PathSearchMethods.GREATER_THAN, "2.9", 3),
        (True, PathSearchMethods.GREATER_THAN, 2.9, "3"),
        (True, PathSearchMethods.GREATER_THAN, "2.9", "3"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2, 4),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2", 4),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2, "4"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2", "4"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2.1, 2.2),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2.1", 2.2),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2.1, "2.2"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2.1", "2.2"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2, 2.1),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2", 2.1),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2, "2.1"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2", "2.1"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2.9, 3),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2.9", 3),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, 2.9, "3"),
        (True, PathSearchMethods.GREATER_THAN_OR_EQUAL, "2.9", "3"),
        (True, PathSearchMethods.LESS_THAN, 4, 2),
        (True, PathSearchMethods.LESS_THAN, "4", 2),
        (True, PathSearchMethods.LESS_THAN, 4, "2"),
        (True, PathSearchMethods.LESS_THAN, "4", "2"),
        (True, PathSearchMethods.LESS_THAN, 4.2, 4.1),
        (True, PathSearchMethods.LESS_THAN, "4.2", 4.1),
        (True, PathSearchMethods.LESS_THAN, 4.2, "4.1"),
        (True, PathSearchMethods.LESS_THAN, "4.2", "4.1"),
        (True, PathSearchMethods.LESS_THAN, 4.2, 4),
        (True, PathSearchMethods.LESS_THAN, "4.2", 4),
        (True, PathSearchMethods.LESS_THAN, 4.2, "4"),
        (True, PathSearchMethods.LESS_THAN, "4.2", "4"),
        (True, PathSearchMethods.LESS_THAN, 4, 3.9),
        (True, PathSearchMethods.LESS_THAN, "4", 3.9),
        (True, PathSearchMethods.LESS_THAN, 4, "3.9"),
        (True, PathSearchMethods.LESS_THAN, "4", "3.9"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4, 2),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4", 2),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4, "2"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4", "2"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4.2, 4.1),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4.2", 4.1),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4.2, "4.1"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4.2", "4.1"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4.2, 4),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4.2", 4),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4.2, "4"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4.2", "4"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4, 3.9),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4", 3.9),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, 4, "3.9"),
        (True, PathSearchMethods.LESS_THAN_OR_EQUAL, "4", "3.9"),
        (True, PathSearchMethods.REGEX, ".+", "a"),
        (True, PathSearchMethods.STARTS_WITH, "p", "parents")
    ])
    def test_search_matches(self, match, method, needle, haystack):
        assert match == Searches.search_matches(method, needle, haystack)

    ###
    # search_anchor
    ###
    def test_search_anchor(self):
        anchor_value = "anchor_name"
        node = ry.scalarstring.PlainScalarString("anchored value", anchor=anchor_value)
        terms = SearchTerms(False, PathSearchMethods.CONTAINS, ".", "name")
        seen_anchors = []
        search_anchors = True
        include_aliases = True
        assert Searches.search_anchor(node, terms, seen_anchors, search_anchors=search_anchors, include_aliases=include_aliases) == AnchorMatches.MATCH



import pytest

from yamlpath.differ.enums.arraydiffopts import ArrayDiffOpts


class Test_differ_enums_arraydiffopts():
	"""Tests for the ArrayDiffOpts enumeration."""

	def test_get_names(self):
		assert ArrayDiffOpts.get_names() == [
			"POSITION",
			"VALUE",
		]

	def test_get_choices(self):
		assert ArrayDiffOpts.get_choices() == [
			"position",
			"value",
		]

	@pytest.mark.parametrize("input,output", [
		("POSITION", ArrayDiffOpts.POSITION),
		("VALUE", ArrayDiffOpts.VALUE),
	])
	def test_from_str(self, input, output):
		assert output == ArrayDiffOpts.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			ArrayDiffOpts.from_str("NO SUCH NAME")



import pytest

from subprocess import run, CalledProcessError

from ruamel.yaml import YAML

from yamlpath.func import unwrap_node_coords
from yamlpath.enums import YAMLValueFormats
from yamlpath.eyaml.enums import EYAMLOutputFormats
from yamlpath.eyaml import EYAMLProcessor
from yamlpath.wrappers import ConsolePrinter
from yamlpath.eyaml.exceptions import EYAMLCommandException

from tests.conftest import requireseyaml, quiet_logger, old_eyaml_keys


@requireseyaml
@pytest.fixture
def eyamldata_f():
    data = """---
aliases:
  - &secretIdentity >
    ENC[PKCS7,MIIBmQYJKoZIhvcNAQcDoIIBijCCAYYCAQAxggEhMIIBHQIBADAFMAACAQEw
    DQYJKoZIhvcNAQEBBQAEggEAUVX4h0PQVnxj5niYRvDPce/TisckEBqkOOcL
    ukGr+AFewRfLQ03zMUcr13jS5w7N6K9TIMPyc0QIvzL82a6jWpNAB7kFD+Ua
    lQcNwFIERYbo3SVn5+r8GTPzS82z59icEgFeL1ChNkL/vRYgys8IJrrJC/uS
    6QQ463hspwF2JyzUF7LM9Jc1EyGuJ1uektj/6jLxnYINrMazC61vb92++2Bk
    eMyFRZyCpJ/0ooHvhtF8ZxlLujPbgaUFCRpCxpXIYOGeTcrqgCZzkU3eUv2r
    PcCqlxHMOjN2SUXBY1pz8ApqErJ9/x0H9lZvD02XYclAMIWb8jouWJA0LaQ0
    Vvji7zBcBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBBVTmGQjl06z3JmnY65
    2STRgDB+D8ySgg5OezkWVRCWXyaei2yeLx4NhKUftXz1G4vbM2rCkFd5Unps
    u30g09oF92k=]
  - &secretPhrase >
    ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEw
    DQYJKoZIhvcNAQEBBQAEggEAXut3j+WbWxcl5Qh857pldvob8b+Av4VwtboA
    14g0slbMCemTtAW8fgpDzmrOGnvPvH5cjcY9FFXdvwa9U02NqosQKV0m7msD
    NirSWcQMGXaOBZqdedIohSHtUKz0QQ0GuEWxFGgXZjGbSwYSzdY6av5zEOFl
    IxGywYD0jLLFm+NkhKxj9wpPyq7qB1JsIqdAbuW33QwgYI8Hp6IBuQCZNFk/
    wnVn9ctInnPhLrvaDt6DN/ikT1d7F7VhteVQzFl3QoL6WW5pcg/vNVp7kvE9
    xJi6eop2BNOuv0Cvm5gJ6OsRX/a/JmA3jZmiweWk5Z3M6OEPpPwnqx1oz3cR
    ES762TBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBBlMyRzGtnxUiel+T+d
    uZFwgCA+DMcg3EllW8xXA7Kc4dhERJLb7VX2mtM9E2s4KrsIyw==]

anchored::secrets: &anchoredSecrets
  exposed_values:
    ident: This identity is in the wild.
    phrase: This phrase is in the wild.
  aliased_values:
    ident: *secretIdentity
    phrase: *secretPhrase
  array_of_array_idents:
    -
      - >
        ENC[PKCS7,MIIBmQYJKoZIhvcNAQcDoIIBijCCAYYCAQAxggEhMIIBHQIBADAFMAACAQEw
        DQYJKoZIhvcNAQEBBQAEggEAxum+Uyt3ETjkaQ9C5PqnpCUVCU6wrUYuVBk+
        PV7t7hayWGrG+dixzUUP9HKbIh6kbVYIwGCpEhMOmJQZ8TLiu/ye+KQzX/CE
        wz4uk7qvv/OvsFiMqmApxcvzNl2Qq7unCScXfngZKPjv4BxAFI1axzsUmxLx
        ChOUSkLMkuIJ5myAw43Sfan9Yx3lk96IoN97gN74ZzXTRGjl3n0zxrHy3obT
        M12f+MZqHuaTnuvksakk32nQ7jGX82QqxX3HChEkzUkKXG5ceS/cFzSTj9QG
        xbYrUXDNq/uviShfVk6tUey76VJAguLlw1ONqRkjonjAz7iR+YIu4RzvPvyJ
        Grz/ezBcBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBDcP+GwmSrNro9UALHI
        FoRFgDCwH91AbH9DOpDMj3HAOlxD2JzAkpy4X9SOZVn6Vht2do38Y1Z02Ccf
        pHj/ObATQ9M=]

aliased::secrets:
  <<: *anchoredSecrets
  novel_values:
    ident: >
      ENC[PKCS7,MIIBmQYJKoZIhvcNAQcDoIIBijCCAYYCAQAxggEhMIIBHQIBADAFMAACAQEw
      DQYJKoZIhvcNAQEBBQAEggEAxum+Uyt3ETjkaQ9C5PqnpCUVCU6wrUYuVBk+
      PV7t7hayWGrG+dixzUUP9HKbIh6kbVYIwGCpEhMOmJQZ8TLiu/ye+KQzX/CE
      wz4uk7qvv/OvsFiMqmApxcvzNl2Qq7unCScXfngZKPjv4BxAFI1axzsUmxLx
      ChOUSkLMkuIJ5myAw43Sfan9Yx3lk96IoN97gN74ZzXTRGjl3n0zxrHy3obT
      M12f+MZqHuaTnuvksakk32nQ7jGX82QqxX3HChEkzUkKXG5ceS/cFzSTj9QG
      xbYrUXDNq/uviShfVk6tUey76VJAguLlw1ONqRkjonjAz7iR+YIu4RzvPvyJ
      Grz/ezBcBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBDcP+GwmSrNro9UALHI
      FoRFgDCwH91AbH9DOpDMj3HAOlxD2JzAkpy4X9SOZVn6Vht2do38Y1Z02Ccf
      pHj/ObATQ9M=]
    phrase: >
      ENC[PKCS7,MIIBygYJKoZIhvcNAQcDoIIBuzCCAbcCAQAxggEhMIIBHQIBADAFMAACAQEw
      DQYJKoZIhvcNAQEBBQAEggEAqaxguhCDS3afSGGsUvQ2bqDXQ71yIRa2sT3N
      1BH8d1fVDEo371jydJhRNg4bAwWjfvOIsuwhkEAtPSycvPNCimvAYz32dybA
      GsQxWo9KuQbmakQ+EDrkxtmS1yxQiUbWk7xw7XercPU8jKSqd3FfpRgKQAKV
      lQeFsCb8Cx/uWg2rS00SZ78LciYeFMwEhx2GYpjmOjdGkMVUKtJhPA24QAeF
      9QSpOmzpIdUho1hxBg9IP5K6HwgNBoYUMhL4xd8hhpRZUn0VSuERvKZAlRyh
      syT4tpPx6MWZw1UXWhPkZg66Fyq+MGJk5Q1zval6Kypqq83SisqsN2z4h1Zx
      WcXJMzCBjAYJKoZIhvcNAQcBMB0GCWCGSAFlAwQBKgQQZRyFQ+GeBvF2A9Yy
      QjUr5oBgwxfr7sa3oJfZNssJy4JlmUhfGVGN5GvxSGHM4zB+2c4lLx9Hk8MZ
      BBdpMsMcD8QqbTFpgOhKTh/pkT9DKP6CUcvSm7oxD7E8RtEyeIP3vH1UlulK
      A0INSZQsnGO+uGIX]
  string_values:
    ident: ENC[PKCS7,MIIBygYJKoZIhvcNAQcDoIIBuzCCAbcCAQAxggEhMIIBHQIBADAFMAACAQEwDQYJKoZIhvcNAQEBBQAEggEADFShQa5BcW0ctXJ6KGiadZYB2hPJrkO6tpOjz3qJzC6zuOrsL80NGOt9njDSEDQhpwvWHaREYJiv4WdBTdRuS2wVkxev/xDMCJrtrTSZ8aFZ2rFy7bkqBx5qiklOtgX2s9jUwzZ1y6YP7HbrBO2d3tO0Df79FvmcgwQOVRUk03BTczbA3xQc1Y50CBoS2d1VE8UxnSUUij3J/tOmugL9QkdSvBIyHwiKiy6brGgwaU1ddGPLMFdRXYN/gpyIbG595YhwTQfDWMp/2gBA7KZf941QJiIvxvq4LoYInNyBK+qTyaVmRRhDTxC4Cs0WfFlkPTkUdGu/GBc32+UDGwDOpTCBjAYJKoZIhvcNAQcBMB0GCWCGSAFlAwQBKgQQ/Ox/sWRra4DBDVh5a4QTD4Bg8v37JkY8S0fNbG/Dq5lRCz7/iQW7c1/f0JfRVIi7qTZfTRElrJo/+/o/SWq6bsj8eUA3UjooR07L0gerjOmd2p1kQQlZ1vEukbF/pnuptS07Gdrs4WlN/6KBIEJ//0Rc]
    phrase: ENC[PKCS7,MIIBqQYJKoZIhvcNAQcDoIIBmjCCAZYCAQAxggEhMIIBHQIBADAFMAACAQEwDQYJKoZIhvcNAQEBBQAEggEAR4DfxkRrYAHFQv97lXvaMyxy2iygBWgXWpKUBskCXbUAU0AZ56dsJS00ibVoRNpBoOkIwVN67G7/z084YA+Oqsg4Tw3NIIek14xChqL9m4ehtv1iMMutPM97wF2Yn2JRs63wKSN4l3nmTp/TFpko5rwj1rKap72mpDwrjJEWwRf4nzcdIzp6a7uWcBUVtG09Cu3VLUtoeAtKsIXVhMAZ2r/ozCSAbIQsFKiRPi2I8fL0ovhnmOXAmuB3eRStMDuGey0vCGYtFvWsmBoXkztIlfHB7/oXUJ5ABgu8D+9JmeXYQA4TjdL6gcQA+cNq9otvorTXnbNLgaRBCGFAqTvMhDBsBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBCw5QXmetCPTaxcpJWAefs+gEDo1hDNBXPFVhtvqqXUzicYZVxDADp2aUo/AdchuG15+8ic7K01aCdL4qkAtyx4HM16Hz0WVYIeiFyUgpCLY1EA]
"""
    yaml = YAML()
    return yaml.load(data)

@pytest.fixture
def force_subprocess_run_cpe(monkeypatch):
    import yamlpath.eyaml.eyamlprocessor as break_module

    def fake_run(*args, **kwargs):
        raise CalledProcessError(42, "bad eyaml")

    monkeypatch.setattr(break_module, "run", fake_run)

@pytest.fixture
def force_no_access(monkeypatch):
    import yamlpath.eyaml.eyamlprocessor as break_module

    def fake_access(*args, **kwargs):
        return False

    monkeypatch.setattr(break_module, "access", fake_access)

class Test_eyaml_EYAMLProcessor():
    def test_find_eyaml_paths(self, quiet_logger, eyamldata_f):
        processor = EYAMLProcessor(quiet_logger, eyamldata_f)
        expected = [
            "aliases[&secretIdentity]",
            "aliases[&secretPhrase]",
            "anchored::secrets.aliased_values.ident",
            "anchored::secrets.aliased_values.phrase",
            "anchored::secrets.array_of_array_idents[0][0]",
            "aliased::secrets.novel_values.ident",
            "aliased::secrets.novel_values.phrase",
            "aliased::secrets.string_values.ident",
            "aliased::secrets.string_values.phrase",
        ]
        actual = []
        for path in processor.find_eyaml_paths():
            actual.append(str(path))

        assert actual == expected

    @requireseyaml
    @pytest.mark.parametrize("yaml_path,compare", [
        ("aliases[&secretIdentity]", "This is not the identity you are looking for."),
        ("aliases[&secretPhrase]", "There is no secret phrase."),
        ("aliases", ["This is not the identity you are looking for.", "There is no secret phrase."]),
        ("(aliases)", ["This is not the identity you are looking for.", "There is no secret phrase."]),
        ("((aliases)*)[0]", "This is not the identity you are looking for."),
    ])
    def test_happy_get_eyaml_values(self, quiet_logger, eyamldata_f, old_eyaml_keys, yaml_path, compare):
        processor = EYAMLProcessor(quiet_logger, eyamldata_f, privatekey=old_eyaml_keys[0], publickey=old_eyaml_keys[1])
        for node in processor.get_eyaml_values(yaml_path, True):
            assert unwrap_node_coords(node) == compare

    @requireseyaml
    @pytest.mark.parametrize("yaml_path,compare,mustexist,output_format", [
        ("aliases[&secretIdentity]", "This is your new identity.", True, EYAMLOutputFormats.STRING),
        ("aliases[&brandNewEntry]", "This key doesn't already exist.", False, EYAMLOutputFormats.BLOCK),
    ])
    def test_happy_set_eyaml_value(self, quiet_logger, eyamldata_f, old_eyaml_keys, yaml_path, compare, mustexist, output_format):
        processor = EYAMLProcessor(quiet_logger, eyamldata_f, privatekey=old_eyaml_keys[0], publickey=old_eyaml_keys[1])

        # Set the test value
        processor.set_eyaml_value(yaml_path, compare, output_format, mustexist)

        # Ensure the new value is encrypted
        encvalue = None
        for encnode in processor.get_nodes(yaml_path):
            encvalue = unwrap_node_coords(encnode)
            break

        assert EYAMLProcessor.is_eyaml_value(encvalue)

    @requireseyaml
    @pytest.mark.parametrize("yaml_path,newval,eoformat,yvformat", [
        ("/aliased::secrets/novel_values/ident", "New, novel, encrypted identity in BLOCK format", EYAMLOutputFormats.BLOCK, YAMLValueFormats.FOLDED),
        ("/aliased::secrets/string_values/ident", "New, novel, encrypted identity in STRING format", EYAMLOutputFormats.STRING, YAMLValueFormats.BARE),
    ])
    def test_preserve_old_blockiness(self, quiet_logger, eyamldata_f, old_eyaml_keys, yaml_path, newval, eoformat, yvformat):
        processor = EYAMLProcessor(quiet_logger, eyamldata_f, privatekey=old_eyaml_keys[0], publickey=old_eyaml_keys[1])
        processor.set_eyaml_value(yaml_path, newval, output=eoformat)

        encvalue = None
        encformat = YAMLValueFormats.DEFAULT
        for encnode in processor.get_nodes(yaml_path):
            encvalue = unwrap_node_coords(encnode)
            encformat = YAMLValueFormats.from_node(encvalue)
            break

        assert EYAMLProcessor.is_eyaml_value(encvalue) and yvformat == encformat

    def test_none_eyaml_value(self):
        assert False == EYAMLProcessor.is_eyaml_value(None)

    @pytest.mark.parametrize("exe", [
        ("/no/such/file/anywhere"),
        ("this-file-does-not-exist"),
        (None),
    ])
    def test_impossible_eyaml_exe(self, exe):
        assert None == EYAMLProcessor.get_eyaml_executable(exe)

    def test_not_can_run_eyaml(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        processor.eyaml = None
        assert False == processor._can_run_eyaml()

    @requireseyaml
    def test_bad_encryption_keys(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        processor.privatekey = "/no/such/file"
        processor.publickey = "/no/such/file"

        with pytest.raises(EYAMLCommandException):
            processor.encrypt_eyaml("test")

    def test_no_encrypt_without_eyaml(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        processor.eyaml = None
        with pytest.raises(EYAMLCommandException):
            processor.encrypt_eyaml("test")

    def test_no_decrypt_without_eyaml(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        processor.eyaml = None
        with pytest.raises(EYAMLCommandException):
            processor.decrypt_eyaml("ENC[...]")

    def test_ignore_already_encrypted_cryps(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        testval = "ENC[...]"
        assert testval == processor.encrypt_eyaml(testval)

    def test_ignore_already_decrypted_cryps(self, quiet_logger):
        processor = EYAMLProcessor(quiet_logger, None)
        testval = "some value"
        assert testval == processor.decrypt_eyaml(testval)

    @requireseyaml
    def test_impossible_decryption(self, quiet_logger, old_eyaml_keys):
        processor = EYAMLProcessor(quiet_logger, None)
        testval = "ENC[...]"
        with pytest.raises(EYAMLCommandException):
            processor.decrypt_eyaml(testval)

    def test_encrypt_calledprocesserror(self, quiet_logger, force_subprocess_run_cpe):
        processor = EYAMLProcessor(quiet_logger, None)
        with pytest.raises(EYAMLCommandException):
            processor.encrypt_eyaml("any value")

    def test_decrypt_calledprocesserror(self, quiet_logger, force_subprocess_run_cpe):
        processor = EYAMLProcessor(quiet_logger, None)
        with pytest.raises(EYAMLCommandException):
            processor.decrypt_eyaml("ENC[...]")

    @requireseyaml
    def test_non_executable(self, old_eyaml_keys, force_no_access):
        assert EYAMLProcessor.get_eyaml_executable(str(old_eyaml_keys[0])) is None



import pytest

from yamlpath.enums import PathSearchMethods


class Test_enums_PathSearchMethods():
	"""Tests for the PathSearchMethods enumeration."""

	@pytest.mark.parametrize("input,output", [
		(PathSearchMethods.CONTAINS, "%"),
		(PathSearchMethods.ENDS_WITH, "$"),
		(PathSearchMethods.EQUALS, "="),
		(PathSearchMethods.STARTS_WITH, "^"),
		(PathSearchMethods.GREATER_THAN, ">"),
		(PathSearchMethods.LESS_THAN, "<"),
		(PathSearchMethods.GREATER_THAN_OR_EQUAL, ">="),
		(PathSearchMethods.LESS_THAN_OR_EQUAL, "<="),
		(PathSearchMethods.REGEX, "=~"),
	])
	def test_str(self, input, output):
		assert output == str(input)

	@pytest.mark.parametrize("input,result", [
		("!", False),
		("%", True),
		("$", True),
		("=", True),
		("^", True),
		(">", True),
		("<", True),
		(">=", True),
		("<=", True),
		("=~", True),
	])
	def test_is_operator(self, input, result):
		assert result == PathSearchMethods.is_operator(input)



"""Make all the YAML Path enumerations available."""
from .anchormatches import AnchorMatches
from .collectoroperators import CollectorOperators
from .includealiases import IncludeAliases
from .pathsearchkeywords import PathSearchKeywords
from .pathsearchmethods import PathSearchMethods
from .pathsegmenttypes import PathSegmentTypes
from .pathseperators import PathSeperators
from .yamlvalueformats import YAMLValueFormats



"""Build this project."""
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="yamlpath",
    description=(
        "Command-line get/set/merge/validate/scan/convert/diff processors for"
        + " YAML/JSON/Compatible data using powerful, intuitive, command-line"
        + " friendly syntax"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    url="https://github.com/wwkimball/yamlpath",
    author="William W. Kimball, Jr., MBA, MSIS",
    author_email="github-yamlpath@kimballstuff.com",
    license="ISC",
    keywords="yaml eyaml json yaml-path diff merge",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "eyaml-rotate-keys = yamlpath.commands.eyaml_rotate_keys:main",
            "yaml-get = yamlpath.commands.yaml_get:main",
            "yaml-paths = yamlpath.commands.yaml_paths:main",
            "yaml-set = yamlpath.commands.yaml_set:main",
            "yaml-merge = yamlpath.commands.yaml_merge:main",
            "yaml-validate = yamlpath.commands.yaml_validate:main",
            "yaml-diff = yamlpath.commands.yaml_diff:main",
        ]
    },
    python_requires=">3.6.0",
    install_requires=[
        "ruamel.yaml>=0.15.96,!=0.17.0,!=0.17.1,!=0.17.2,!=0.17.5,!=0.17.18,<=0.17.21",
        "python-dateutil<=3"
    ],
    tests_require=[
        "pytest",
        "pytest-cov",
        "pytest-console-scripts",
    ],
    include_package_data=True,
    zip_safe=False
)



"""
Enable users to merge YAML/Compatible files.

Due to the complexities of merging, users are given deep control over the merge
operation via both default behaviors as well as per YAML Path behaviors.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
import sys
import argparse
import json
from os import access, R_OK, remove
from os.path import isfile, exists
from shutil import copy2
from typing import List, Tuple

from ruamel.yaml import YAML

from yamlpath import __version__ as YAMLPATH_VERSION
from yamlpath.common import Parsers
from yamlpath.merger.enums import (
    AnchorConflictResolutions,
    AoHMergeOpts,
    ArrayMergeOpts,
    HashMergeOpts,
    MultiDocModes,
    OutputDocTypes,
    SetMergeOpts,
)
from yamlpath.merger.exceptions import MergeException
from yamlpath.merger import Merger, MergerConfig
from yamlpath.exceptions import YAMLPathException

from yamlpath.wrappers import ConsolePrinter

def processcli():
    """Process command-line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Merges two or more single- or multi-document"
                    " YAML/JSON/Compatible documents\ntogether, including"
                    " complex data provided via STDIN.",
        epilog="""
configuration file:
  The CONFIG file is an INI file with up to three sections:
  [defaults] Sets equivalents of --anchors|-a, --arrays|-A, --hashes|-H, and
             --aoh|-O.
  [rules]    Each entry is a YAML Path assigning --arrays|-A, --hashes|-H,
             or --aoh|-O for precise nodes.
  [keys]     Wherever --aoh=DEEP (or -O deep), each entry is treated as a
             record with an identity key.  In order to match RHS records to
             LHS records, a key must be known and is identified on a YAML
             Path basis via this section.  Where not specified, the first
             attribute of the first record in the Array-of-Hashes is presumed
             the identity key for all records in the set.

input files:
  The left-to-right order of YAML_FILEs is significant.  Except when this
  behavior is deliberately altered by your options, data from files on the
  right overrides data in files to their left.

  Only one input file may be the - pseudo-file (read from STDIN).  When no
  YAML_FILEs are provided, - will be inferred as long as you are running
  this program without a TTY (unless you set --nostdin|-S).

  Any file, including input from STDIN, may be a multi-document YAML, JSON,
  or compatible file.

For more information about YAML Paths, please visit
https://github.com/wwkimball/yamlpath/wiki.

To report issues with this tool or to request enhancements, please visit
https://github.com/wwkimball/yamlpath/issues.
""")

    parser.add_argument("-V", "--version", action="version",
                        version="%(prog)s " + YAMLPATH_VERSION)

    parser.add_argument(
        "-c", "--config", help=(
            "INI syle configuration file for YAML Path specified\n"
            "merge control options"))

    parser.add_argument(
        "-a", "--anchors",
        choices=[l.lower() for l in AnchorConflictResolutions.get_names()],
        type=str.lower,
        help=(
            "means by which Anchor name conflicts are resolved\n"
            "(overrides [defaults]anchors set via --config|-c and\n"
            "cannot be overridden by [rules] because Anchors apply\n"
            "to the whole file); default=stop"))
    parser.add_argument(
        "-A", "--arrays",
        choices=[l.lower() for l in ArrayMergeOpts.get_names()],
        type=str.lower,
        help=(
            "default means by which Arrays are merged together\n"
            "(overrides [defaults]arrays but is overridden on a\n"
            "YAML Path basis via --config|-c); default=all"))
    parser.add_argument(
        "-E", "--sets",
        choices=[l.lower() for l in SetMergeOpts.get_names()],
        type=str.lower,
        help=(
            "default means by which Sets are merged together\n"
            "(overrides [defaults]sets but is overridden on a\n"
            "YAML Path basis via --config|-c); default=unique"))
    parser.add_argument(
        "-H", "--hashes",
        choices=[l.lower() for l in HashMergeOpts.get_names()],
        type=str.lower,
        help=(
            "default means by which Hashes are merged together\n"
            "(overrides [defaults]hashes but is overridden on a\n"
            "YAML Path basis in [rules] set via --config|-c);\n"
            "default=deep"))
    parser.add_argument(
        "-O", "--aoh",
        choices=[l.lower() for l in AoHMergeOpts.get_names()],
        type=str.lower,
        help=(
            "default means by which Arrays-of-Hashes are merged\n"
            "together (overrides [defaults]aoh but is overridden on\n"
            "a YAML Path basis in [rules] set via --config|-c);\n"
            "default=all"))

    parser.add_argument(
        "-m", "--mergeat",
        metavar="YAML_PATH",
        default="/",
        help=(
            "YAML Path indicating where in left YAML_FILE the right\n"
            "YAML_FILE content is to be merged; default=/"))

    output_doc_group = parser.add_mutually_exclusive_group()
    output_doc_group.add_argument(
        "-o", "--output",
        help=(
            "write the merged result to the indicated nonexistent\n"
            "file"))
    output_doc_group.add_argument(
        "-w", "--overwrite",
        help=(
            "write the merged result to the indicated file; will\n"
            "replace the file when it already exists"))

    parser.add_argument(
        "-b", "--backup", action="store_true",
        help=(
            "save a backup OVERWRITE file with an extra .bak\n"
            "file-extension; applies only to OVERWRITE"))

    parser.add_argument(
        "-D", "--document-format",
        choices=[l.lower() for l in OutputDocTypes.get_names()],
        type=str.lower,
        default="auto",
        help=(
            "force the merged result to be presented in one of the\n"
            "supported formats or let it automatically match the\n"
            "known file-name extension of OUTPUT|OVERWRITE (when\n"
            "provided), or match the type of the first document;\n"
            "default=auto"))

    parser.add_argument(
        "-M", "--multi-doc-mode",
        choices=[l.lower() for l in MultiDocModes.get_names()],
        type=str.lower,
        default="condense_all",
        help=(
            "control how multi-document files and streams are\n"
            "merged together, with or without condensing them as\n"
            "part of the merge"))

    parser.add_argument(
        "-l", "--preserve-lhs-comments", action="store_true",
        help=(
            "while all comments are normally dicarded during a\n"
            "merge, this option will attempt to preserve\n"
            "comments in the left-most YAML_FILE; may produce\n"
            "unexpected comment-to-data associations or\n"
            "spurious new-lines and all other document comments\n"
            "are still discarded"))

    parser.add_argument(
        "-S", "--nostdin", action="store_true",
        help=(
            "do not implicitly read from STDIN, even when there are\n"
            "no - pseudo-files in YAML_FILEs with a non-TTY session"))

    noise_group = parser.add_mutually_exclusive_group()
    noise_group.add_argument(
        "-d", "--debug", action="store_true",
        help="output debugging details")
    noise_group.add_argument(
        "-v", "--verbose", action="store_true",
        help="increase output verbosity")
    noise_group.add_argument(
        "-q", "--quiet", action="store_true",
        help=(
            "suppress all output except errors (implied when\n"
            "-o|--output is not set)"))

    parser.add_argument(
        "yaml_files", metavar="YAML_FILE", nargs="*",
        help=(
            "one or more YAML files to merge, order-significant;\n"
            "omit or use - to read from STDIN"))
    return parser.parse_args()

# pylint: disable=too-many-branches
def validateargs(args, log):
    """Validate command-line arguments."""
    has_errors = False

    # There must be at least one input file or stream
    input_file_count = len(args.yaml_files)
    if (input_file_count == 0 and (
            sys.stdin.isatty()
            or args.nostdin)
    ):
        has_errors = True
        log.error(
            "There must be at least one YAML_FILE or STDIN document.")

    # There can be only one -
    pseudofile_count = 0
    for infile in args.yaml_files:
        if infile.strip() == '-':
            pseudofile_count += 1
    if pseudofile_count > 1:
        has_errors = True
        log.error("Only one YAML_FILE may be the - pseudo-file.")

    # When set, the configuration file must be a readable file
    if args.config and not (
            isfile(args.config)
            and access(args.config, R_OK)
    ):
        has_errors = True
        log.error(
            "INI style configuration file is not readable:  {}"
            .format(args.config))

    # When set, the output file must not already exist
    if args.output:
        if exists(args.output):
            has_errors = True
            log.error("Output file already exists:  {}".format(args.output))
    elif args.overwrite:
        if exists(args.overwrite):
            log.warning(
                "Output file exists and will be overwritten:  {}"
                .format(args.overwrite))
    else:
        # When dumping the document to STDOUT, mute all non-errors except when
        # forced.
        force_verbose = args.verbose
        force_debug = args.debug
        if not (force_verbose or force_debug):
            args.quiet = True
            args.verbose = False
            args.debug = False

    # When set, backup applies only to OVERWRITE
    if args.backup and not args.overwrite:
        has_errors = True
        log.error("The --backup|-b option applies only to OVERWRITE files.")

    if has_errors:
        sys.exit(1)

def write_output_document(
    args: argparse.Namespace, log: ConsolePrinter, yaml_editor: YAML,
    docs: List[Merger]
) -> None:
    """Save a backup of the overwrite file, if requested."""
    if args.backup:
        backup_file = args.overwrite + ".bak"
        log.verbose(
            "Saving a backup of {} to {}."
            .format(args.overwrite, backup_file))
        if exists(backup_file):
            remove(backup_file)
        copy2(args.overwrite, backup_file)

    document_is_json = (
        docs[0].prepare_for_dump(yaml_editor, args.output)
        is OutputDocTypes.JSON)

    dumps = []
    for doc in docs:
        doc.prepare_for_dump(yaml_editor, args.output)
        dumps.append(doc.data)

    if args.output:
        with open(args.output, 'w', encoding='utf-8') as out_fhnd:
            if document_is_json:
                if len(dumps) > 1:
                    for dump in dumps:
                        print(
                            json.dumps(Parsers.jsonify_yaml_data(dump)),
                            file=out_fhnd)
                else:
                    json.dump(Parsers.jsonify_yaml_data(dumps[0]), out_fhnd)
            else:
                if len(dumps) > 1:
                    yaml_editor.explicit_end = True  # type: ignore
                    yaml_editor.dump_all(dumps, out_fhnd)
                else:
                    yaml_editor.dump(dumps[0], out_fhnd)
    else:
        if document_is_json:
            if len(dumps) > 1:
                for dump in dumps:
                    print(json.dumps(Parsers.jsonify_yaml_data(dump)))
            else:
                json.dump(Parsers.jsonify_yaml_data(dumps[0]), sys.stdout)
        else:
            if len(dumps) > 1:
                yaml_editor.explicit_end = True  # type: ignore
                yaml_editor.dump_all(dumps, sys.stdout)
            else:
                yaml_editor.dump(dumps[0], sys.stdout)

def get_doc_mergers(
    log: ConsolePrinter, yaml_editor: YAML, config: MergerConfig,
    yaml_file: str
) -> Tuple[List[Merger], bool]:
    """Create a list of Mergers, one for each source document."""
    docs_loaded = True
    if yaml_file != "-" and not isfile(yaml_file):
        log.error("Not a file:  {}".format(yaml_file))
        return ([], False)

    doc_mergers: List[Merger] = []
    for (yaml_data, doc_loaded) in Parsers.get_yaml_multidoc_data(
        yaml_editor, log, yaml_file
    ):
        if not doc_loaded:
            # An error message has already been logged
            doc_mergers.clear()
            docs_loaded = False
            break

        doc_mergers.append(Merger(log, yaml_data, config))

    return (doc_mergers, docs_loaded)

def merge_condense_all(
    log: ConsolePrinter, lhs_docs: List[Merger], rhs_docs: List[Merger]
) -> int:
    """Condense LHS and RHS multi-docs together into one."""
    return_state = 0
    lhs_prime = lhs_docs[0]
    if len(lhs_docs) > 1:
        for lhs_doc in lhs_docs[1:]:
            try:
                lhs_prime.merge_with(lhs_doc.data)
            except MergeException as mex:
                log.error(mex)
                return_state = 11
            except YAMLPathException as yex:
                log.error(yex)
                return_state = 12

    # With all subdocs merged into the first, eliminate all subdocs
    for i in reversed(range(1, len(lhs_docs))):
        del lhs_docs[i]

    # Merge every RHS doc into the prime LHS doc
    for rhs_doc in rhs_docs:
        try:
            lhs_prime.merge_with(rhs_doc.data)
        except MergeException as mex:
            log.error(mex)
            return_state = 13
        except YAMLPathException as yex:
            log.error(yex)
            return_state = 14

    return return_state

def merge_across(
    log: ConsolePrinter, lhs_docs: List[Merger], rhs_docs: List[Merger]
) -> int:
    """Condense LHS and RHS multi-docs together into one."""
    return_state = 0
    lhs_len = len(lhs_docs)
    rhs_len = len(rhs_docs)
    max_len = lhs_len if lhs_len > rhs_len else rhs_len
    lhs_limit = lhs_len - 1
    rhs_limit = rhs_len - 1
    for i in range(0, max_len):
        if i > rhs_limit:
            break
        if i > lhs_limit:
            lhs_docs.append(rhs_docs[i])
            continue
        try:
            lhs_docs[i].merge_with(rhs_docs[i].data)
        except MergeException as mex:
            log.error(mex)
            return_state = 31
            break
        except YAMLPathException as yex:
            log.error(yex)
            return_state = 32
            break

    return return_state

def merge_matrix(
    log: ConsolePrinter, lhs_docs: List[Merger], rhs_docs: List[Merger]
) -> int:
    """Condense LHS and RHS multi-docs together into one."""
    return_state = 0
    for lhs_doc in lhs_docs:
        for rhs_doc in rhs_docs:
            try:
                lhs_doc.merge_with(rhs_doc.data)
            except MergeException as mex:
                log.error(mex)
                return_state = 41
                break
            except YAMLPathException as yex:
                log.error(yex)
                return_state = 42
                break

    return return_state

def merge_docs(
    log: ConsolePrinter, yaml_editor: YAML, config: MergerConfig,
    lhs_docs: List[Merger], rhs_file: str
) -> int:
    """Merge RHS into LHS."""
    return_state = 0
    merge_mode = config.get_multidoc_mode()
    (rhs_docs, rhs_loaded) = get_doc_mergers(
        log, yaml_editor, config, rhs_file)
    if not rhs_loaded:
        # Failed to load any RHS documents
        return 3

    if merge_mode is MultiDocModes.CONDENSE_ALL:
        return_state = merge_condense_all(log, lhs_docs, rhs_docs)

    elif merge_mode is MultiDocModes.MERGE_ACROSS:
        return_state = merge_across(log, lhs_docs, rhs_docs)

    elif merge_mode is MultiDocModes.MATRIX_MERGE:
        return_state = merge_matrix(log, lhs_docs, rhs_docs)

    return return_state

def main():
    """Main code."""
    args = processcli()
    log = ConsolePrinter(args)
    validateargs(args, log)

    # For the remainder of processing, overwrite overwrites output
    if args.overwrite:
        args.output = args.overwrite

    # Merge all input files
    yaml_editor = Parsers.get_yaml_editor()
    merge_config = MergerConfig(log, args)
    exit_state = 0
    consumed_stdin = False
    mergers: List[Merger] = []
    merge_count = 0
    for yaml_file in args.yaml_files:
        if yaml_file.strip() == '-':
            consumed_stdin = True

        log.debug(
            "yaml_merge::main:  Processing file, {}".format(
                "STDIN" if yaml_file.strip() == "-" else yaml_file))

        if len(mergers) < 1:
            (mergers, mergers_loaded) = get_doc_mergers(
                log, yaml_editor, merge_config, yaml_file)
            if not mergers_loaded:
                exit_state = 4
                break
        else:
            # Merge RHS into LHS
            exit_state = merge_docs(
                log, yaml_editor, merge_config, mergers, yaml_file)
            if not exit_state == 0:
                break
            merge_count += 1

    # Check for a waiting STDIN document
    if (exit_state == 0
        and not consumed_stdin
        and not args.nostdin
        and not sys.stdin.isatty()
    ):
        exit_state = merge_docs(log, yaml_editor, merge_config, mergers, "-")
        merge_count += 1

    # When no merges have occurred, check for a single-doc merge request
    if (exit_state == 0
        and merge_count == 0
        and merge_config.get_multidoc_mode() is MultiDocModes.CONDENSE_ALL
    ):
        exit_state = merge_condense_all(log, mergers, [])

    # Output the final document
    if exit_state == 0:
        write_output_document(args, log, yaml_editor, mergers)

    sys.exit(exit_state)

if __name__ == "__main__":
    main()  # pragma: no cover



"""
Implements the IncludeAliases enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto


class IncludeAliases(Enum):
    """
    When performing a search, YAML Anchors and Aliases can be evaluated.

    Whether either are is dictated by these options:

    `ANCHORS_ONLY`
        Only anchors are evaluated.

    `INCLUDE_KEY_ALIASES`
        Anchors and key aliases are evaluated.

    `INCLUDE_VALUE_ALIASES`
        Anchors and value aliases are evaluated.

    `INCLUDE_ALL_ALIASES`
        Anchors and all aliases are evaluated.
    """

    ANCHORS_ONLY = auto()
    INCLUDE_KEY_ALIASES = auto()
    INCLUDE_VALUE_ALIASES = auto()
    INCLUDE_ALL_ALIASES = auto()



"""
YAML Path Collector segment terms.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from yamlpath.enums import CollectorOperators


class CollectorTerms:
    """YAML Path Collector segment terms."""

    def __init__(
        self, expression: str,
        operation: CollectorOperators = CollectorOperators.NONE
    ) -> None:
        """
        Instantiate a Collector Term.

        Parameters:
        1. expression (str) The YAML Path being collected
        2. operation (CollectorOperators) The operation for this Collector,
           relative to its subsequent peer Collector
        """
        self._expression: str = expression
        self._operation: CollectorOperators = operation

    def __str__(self) -> str:
        """Get the String rendition of this Collector Term."""
        operator: str = str(self.operation)
        return "{}({})".format(operator, self.expression)

    @property
    def operation(self) -> CollectorOperators:
        """
        Get the operation for this Collector.

        This indicates whether its results are independent, added to the prior
        Collector, or removed from the prior Collector.
        """
        return self._operation

    @property
    def expression(self) -> str:
        """Get the Collector expression, which is a stringified YAML Path."""
        return self._expression



import pytest

from yamlpath.enums import PathSearchMethods
from yamlpath.path import SearchTerms

class Test_path_SearchTerms():
	"""Tests for the SearchTerms class."""

	@pytest.mark.parametrize("invert,method,attr,term,output", [
		(False, PathSearchMethods.CONTAINS, "abc", "b", "[abc%b]"),
		(True, PathSearchMethods.REGEX, "abc", "^abc$", "[abc!=~/^abc$/]"),
	])
	def test_str(self, invert, method, attr, term, output):
		assert output == str(SearchTerms(invert, method, attr, term))

	# Disabled until Python matures enough to permit classes and types to play
	# nicely together...
	# def test_from_path_segment_attrs(self):
	# 	from yamlpath.types import PathAttributes
	# 	st = SearchTerms(False, PathSearchMethods.EQUALS, ".", "key")
	# 	assert str(st) == str(SearchTerms.from_path_segment_attrs(st))

	# 	with pytest.raises(AttributeError):
	# 		_ = SearchTerms.from_path_segment_attrs("nothing-to-see-here")



import pytest

from yamlpath.exceptions import YAMLPathException


class Test_exceptions_YAMLPathException():
	def test_str_segmentless(self):
		with pytest.raises(YAMLPathException) as ex:
			raise YAMLPathException("Test message", "test")
		assert str(ex.value) == "Test message, 'test'."

	def test_str_segmentfull(self):
		with pytest.raises(YAMLPathException) as ex:
			raise YAMLPathException("Test message", "test", "st")
		assert str(ex.value) == "Test message at 'st' in 'test'."



"""
Implements the MultiDocModes enumeration.

Copyright 2021 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class MultiDocModes(Enum):
    """
    Supported means of merging multi-document content.

    Options include:

    `CONDENSE_ALL`
        Merge all multi-documents up into single documents during the merge.

    `MERGE_ACROSS`
        Condence no multi-documents; rather, only merge documents "across" from
        right to left.

    `MATRIX_MERGE`
        Condence no multi-documents; rather, merge every RHS document into
        every LHS document.
    """

    CONDENSE_ALL = auto()
    MERGE_ACROSS = auto()
    MATRIX_MERGE = auto()

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all upper-cased entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in MultiDocModes]

    @staticmethod
    def get_choices() -> List[str]:
        """
        Get all entry names with symbolic representations for this enumeration.

        All returned entries are lower-cased.

        Parameters:  N/A

        Returns:  (List[str]) Lower-case names and symbols from this
            enumeration

        Raises:  N/A
        """
        names = [l.lower() for l in MultiDocModes.get_names()]
        choices = list(set(names))
        choices.sort()
        return choices

    @staticmethod
    def from_str(name: str) -> "MultiDocModes":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (MultiDocModes) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values
        """
        check: str = str(name).upper()
        if check in MultiDocModes.get_names():
            return MultiDocModes[check]
        raise NameError(
            "MultiDocModes has no such item:  {}"
            .format(name))



"""
Implement NodeCoords.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
from typing import Any, List, Optional, Type

from yamlpath.types import AncestryEntry, PathSegment
from yamlpath import YAMLPath

class NodeCoords:
    """
    Wrap a node's data along with its relative coordinates within its DOM.

    A node's "coordinates" includes these properties:
    1. Reference to the node itself,
    2. Immediate parent node of the wrapped node,
    3. Index or Key of the node within its immediate parent

    Additional, optional data can be wrapped along with the node's coordinates
    to facilitate other specific operations upon the node/DOM.  See the
    `__init__` method for details.
    """

    # pylint: disable=locally-disabled,too-many-arguments
    def __init__(
        self, node: Any, parent: Any, parentref: Any,
        path: Optional[YAMLPath] = None,
        ancestry: Optional[List[AncestryEntry]] = None,
        path_segment: Optional[PathSegment] = None
    ) -> None:
        """
        Initialize a new NodeCoords.

        Positional Parameters:
        1. node (Any) Reference to the ruamel.yaml DOM data element
        2. parent (Any) Reference to `node`'s immediate DOM parent
        3. parentref (Any) The `list` index or `dict` key which indicates where
           within `parent` the `node` is located
        4. path (YAMLPath) The YAML Path for this node, as reported by its
           creator process
        5. ancestry (List[AncestryEntry]) Stack of AncestryEntry (parent,
           parentref) tracking the hierarchical ancestry of this node through
           its parent document
        6. path_segment (PathSegment) The YAML Path segment which most directly
           caused the generation of this NodeCoords

        Returns: N/A

        Raises:  N/A
        """
        self.node: Any = node
        self.parent: Any = parent
        self.parentref: Any = parentref
        self.path: Optional[YAMLPath] = path
        self.ancestry: List[AncestryEntry] = ([]
                                              if ancestry is None
                                              else ancestry)
        self.path_segment: Optional[PathSegment] = path_segment

    def __str__(self) -> str:
        """Get a String representation of this object."""
        return str(self.node)

    def __repr__(self) -> str:
        """
        Generate an eval()-safe representation of this object.

        Assumes all of the ruamel.yaml components are similarly safe.
        """
        return ("{}('{}', '{}', '{}')".format(
            self.__class__.__name__, self.node, self.parent,
            self.parentref))

    def __gt__(self, rhs: "NodeCoords") -> Any:
        """Indicate whether this node's data is greater-than another's."""
        if self.node is None or rhs.node is None:
            return False
        return self.node > rhs.node

    def __lt__(self, rhs: "NodeCoords") -> Any:
        """Indicate whether this node's data is less-than another's."""
        if self.node is None or rhs.node is None:
            return False
        return self.node < rhs.node

    @property
    def unwrapped_node(self) -> Any:
        """Unwrap the data, no matter how deeply nested it may be."""
        return NodeCoords.unwrap_node_coords(self)

    @property
    def deepest_node_coord(self) -> "NodeCoords":
        """Get the deepest wrapped NodeCoord contained within."""
        return NodeCoords._deepest_node_coord(self)

    def wraps_a(self, compare_type: Type) -> bool:
        """Indicate whether the wrapped node is of a given data-type."""
        if compare_type is None:
            return self.unwrapped_node is None
        return isinstance(self.unwrapped_node, compare_type)

    @staticmethod
    def _deepest_node_coord(node: "NodeCoords") -> "NodeCoords":
        """Get the deepest nested NodeCoord."""
        if (not isinstance(node, NodeCoords)
            or not isinstance(node.node, NodeCoords)
        ):
            return node

        return NodeCoords._deepest_node_coord(node.node)

    @staticmethod
    def unwrap_node_coords(data: Any) -> Any:
        """
        Recursively strips all DOM tracking data off of a NodeCoords wrapper.

        Parameters:
        1. data (Any) the source data to strip.

        Returns:  (Any) the stripped data.
        """
        if isinstance(data, NodeCoords):
            return NodeCoords.unwrap_node_coords(data.node)

        if isinstance(data, list):
            stripped_nodes = []
            for ele in data:
                stripped_nodes.append(NodeCoords.unwrap_node_coords(ele))
            return stripped_nodes

        return data



"""
Implements the YAMLValueFormats enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
import datetime
from enum import Enum, auto
from typing import Any, List

from ruamel.yaml.scalarstring import (
    PlainScalarString,
    DoubleQuotedScalarString,
    SingleQuotedScalarString,
    FoldedScalarString,
    LiteralScalarString,
)
from ruamel.yaml.scalarbool import ScalarBoolean
from ruamel.yaml.scalarfloat import ScalarFloat
from ruamel.yaml.scalarint import ScalarInt
# pylint: disable=wrong-import-position,ungrouped-imports
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
        AnchoredDate,
    )  # type: ignore
else:                                         # pragma: no cover
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    AnchoredDate = AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp
    # From whence shall come AnchoredDate?
# pylint: enable=wrong-import-position,ungrouped-imports


class YAMLValueFormats(Enum):
    """
    Supported representation formats for YAML values.

    These include:

    `BARE`
        The value is written as-is, when possible, with neither demarcation nor
        reformatting.  The YAML parser may convert the format to something else
        if it deems necessary.

    `BOOLEAN`
        The value is written as a bare True or False.

    `DATE`
        The value is written as a bare ISO8601 date without a time component.

    `DEFAULT`
        The value is written in whatever format is deemed most appropriate.

    `DQUOTE`
        The value is demarcated via quotation-marks (").

    `FLOAT`
        The value is written as a bare floating-point decimal.

    `FOLDED`
        An otherwise long single-line string is written as a multi-line value
        which YAML data parsers can read back as the original single-line
        string.

    `INT`
        The value is written as a bare integer number with no fractional
        component.

    `LITERAL`
        A multi-line string is written as-is, preserving newline characters and
        any other white-space.

    `SQUOTE`
        The value is demarcated via apostrophes (').

    `TIMESTAMP`
        The value is a timestamp per the supported syntax of ISO8601 by
        http://yaml.org/type/timestamp.html.
    """

    BARE = auto()
    BOOLEAN = auto()
    DATE = auto()
    DEFAULT = auto()
    DQUOTE = auto()
    FLOAT = auto()
    FOLDED = auto()
    INT = auto()
    LITERAL = auto()
    SQUOTE = auto()
    TIMESTAMP = auto()

    @staticmethod
    def get_names() -> List[str]:
        """
        Return all entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in YAMLValueFormats]

    @staticmethod
    def from_str(name: str) -> "YAMLValueFormats":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (YAMLValueFormats) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values.
        """
        check: str = str(name).upper()
        if check in YAMLValueFormats.get_names():
            return YAMLValueFormats[check]
        raise NameError(
            "YAMLValueFormats has no such item:  {}"
            .format(name))

    @staticmethod
    def from_node(node: Any) -> "YAMLValueFormats":
        """
        Identify the best matching enumeration value from a sample data node.

        Will return YAMLValueFormats.DEFAULT if the node is None or its best
        match cannot be determined.

        Parameters:
            1. node (Any) The node to type

        Returns:  (YAMLValueFormats) one of the enumerated values

        Raises:  N/A
        """
        best_type: YAMLValueFormats = YAMLValueFormats.DEFAULT
        if node is None:
            return best_type

        node_type: type = type(node)
        if node_type is FoldedScalarString:
            best_type = YAMLValueFormats.FOLDED
        elif node_type is LiteralScalarString:
            best_type = YAMLValueFormats.LITERAL
        elif node_type is DoubleQuotedScalarString:
            best_type = YAMLValueFormats.DQUOTE
        elif node_type is SingleQuotedScalarString:
            best_type = YAMLValueFormats.SQUOTE
        elif node_type is PlainScalarString:
            best_type = YAMLValueFormats.BARE
        elif node_type is ScalarBoolean:
            best_type = YAMLValueFormats.BOOLEAN
        elif node_type is ScalarFloat:
            best_type = YAMLValueFormats.FLOAT
        elif node_type is ScalarInt:
            best_type = YAMLValueFormats.INT
        elif node_type is AnchoredDate or node_type is datetime.date:
            best_type = YAMLValueFormats.DATE
        elif node_type is AnchoredTimeStamp or node_type is datetime.datetime:
            best_type = YAMLValueFormats.TIMESTAMP

        return best_type
